#define NEW gsd_clipboard_manager_new
#define START gsd_clipboard_manager_start
#define STOP gsd_clipboard_manager_stop
#define MANAGER GsdClipboardManager
#include "gsd-clipboard-manager.h"

#include "daemon-skeleton-gtk.h"
